// This version of the Twirthie source code file is identical to the
// versions bundled with the .class files, but has lots of comments.

// Twirthie is a console-mode random sentence generator.
// It uses a grammar similar to the author's intrusive, repetitive
// thoughts (a positive symptom of schizophrenia).

import org.json.simple.JSONObject;
import org.json.simple.JSONArray;
import org.json.simple.parser.ParseException;
import org.json.simple.parser.JSONParser;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.lang.*;

class twirt {

    static final char SYMCH = '$';
    static final char NUMLITCH = '#';
    static final int lineSize = 65;
    static final int maxSentenceCount = 50;
    static boolean verbose = false;

    public static void main(String[] args){
	
      JSONParser parser = new JSONParser();
      Object obj, tokobj;
      JSONArray array;
      GrmObj grmObj;
      String sentence, paragraph;
      String filePath = "";  // "/home/mike/plictalk/droptagger/src/";
      String fileName = "twirt.grm";
      String fullFileName = filePath + fileName;
      String sGrammar;
      int symCount, ruleCount, tokenCount, optCount, i, j, k, m;
      int pct;
      int errcode = 0;
      int sentenceCount = 1;
      JSONObject symdef;
      String sym, token, subToken, msg;
      JSONArray rules, rule, opt;
      int rulesFound = 0;
      int optRulesFound = 0;
      int strTokensFound = 0;
      String sRule = "";
      boolean optFound;

      // First command-line argument equals no. of sentences to
      // generate (default = 1)
      // Use 0 to echo your input
      if (args.length > 0) {
         try {
            i = Integer.parseInt(args[0]);
         } catch (NumberFormatException e) {
            i = 1;
         }
         if ((i > 1) && (args.length == 1)) {
            sentenceCount = Math.min(i, maxSentenceCount);
         }
         verbose = (i == 0);
      }
      try {
         sGrammar = getGrm(fullFileName); // grammar in oddball format
         obj = parser.parse(sGrammar);  // parse grammar as JSON array
         if (verbose)
            printGrm(sGrammar, obj);  
         array = (JSONArray)obj;
         symCount = array.size();
         println(symCount + " symdefs in grammar");
         println("");
         for (i = 0; i < symCount; i++) {
            symdef = (JSONObject)array.get(i);
            sym = (String)symdef.get("sym");
            rules = (JSONArray)symdef.get("rules");
            ruleCount = rules.size();
            if (ruleCount <= 0) {  // can't have sym hdr w/ no rules
                errcode = 1;
                break;
            }
            println(sym);
            for (j = 0; j < ruleCount; j++) {
                rulesFound++;
                rule = (JSONArray)rules.get(j);
                tokenCount = rule.size();
                if (tokenCount < 2) {  // rule needs >= 2 tokens
                    errcode = 2;
                    break;
                }
                sRule = "  ";
                optFound = false;
                for (k = 0; k < tokenCount; k++) {
                    if (k == 0) {
                        // first token is percentage (probability)
                        pct = parsePct(rule.get(k));
                        if (pct < -1) {
                            errcode = 3;
                            break;
                        }
                        token = "" + pct;
                    }
                    else {
                        tokobj = rule.get(k);
                        if (tokobj instanceof JSONArray) {
                            // optional block in square brackets
                            optFound = true;
                            opt = (JSONArray)tokobj;
                            optCount = opt.size();
                            if (optCount < 2) {
                                errcode = 4;
                                break;
                            }
                            token = "";
                            for (m = 0; m < optCount; m++) {
                                if (m == 0) {
                                    // first token is percentage
                                    pct = parsePct(opt.get(m));
                                    if (pct < -1) {
                                        errcode = 3;
                                        break;
                                    }
                                    token = "[" + pct;
                                }
                                else {
                                    subToken = (String)opt.get(m);
                                    token = token + " " + subToken;
                                    strTokensFound++;
                                }
                            }
                            if (errcode != 0) 
                                break; 
                            token = token + "]";
                        }
                        else {
                            // normal string token
                            token = (String)tokobj;
                            strTokensFound++;
                        }
                    }
                    sRule = sRule + token + " ";
                }
                println(sRule);
                if (errcode != 0)
                    break;
                if (optFound)
                    optRulesFound++;
            }
            println("");
            if (errcode != 0)
                break;
         }
         if (errcode == 0) {
            // generate random sentence(s):
            grmObj = new GrmObj(array);
            paragraph = "";
            for (i = 0; i < sentenceCount; i++) {
                sentence = grmObj.getSentence();
                paragraph = grmObj.concTrim(paragraph, sentence);
            }
            if (!verbose) 
                System.out.println("");
            printWordWrap(paragraph);
            System.out.println("");
         }
         println(rulesFound + " rules found");
         // count rules w/ at least one optional block:
         println(optRulesFound + " rules with opt. blks. found");
         println(strTokensFound + " string tokens found");
         println("");
         msg = "";
         switch (errcode) {
            case 0:  // no errors encountered
                break;
            case 1:
                msg = "empty symdef encountered";
                break;
            case 2:
                msg = "no string tokens found within a rule def";
                break;
            case 3:
                msg = "percentage value must be between -1 and 100";
                println('{'+sRule+'}');
                break;
            case 4:
                msg = "no string tokens found within an optional sub-rule";
                break;
            default:
                msg = "internal error - bad errcode";
         }
         if (msg.length() > 0) {
            System.out.println("Error detected (code = " + errcode +
                "): " + msg);
            System.out.println("");
         }
         
      }catch(ParseException pe){
		
         System.out.println("position: " + pe.getPosition());
         System.out.println(pe);
      }
    }
    
    static void println(String msg) {
        if (verbose) {
            System.out.println(msg);
        }
    }
    
    static void outln(String msg) {
        println('{'+msg+'}');
    }
     
    static String getGrm(String fileName) {  // parse the oddball fmt.
		BufferedReader br = null;
		String rtnval = "[";
		String outbuf = "";
		char leadch;
		char sp = ' ';
		int lineCount = 0;

		try {
			String inbuf;
			br = new BufferedReader(new FileReader(fileName));
			while ((inbuf = br.readLine()) != null) {
			    if (inbuf.trim().length() == 0)
			        continue;
				println(inbuf);
				outbuf = "";
				leadch = inbuf.charAt(0);
    		    inbuf = inbuf.trim();
            	if (leadch == sp) {
                    // indented line is a rule
				    outbuf = "[" + parseRuleGrm(inbuf) + "],";
				}
				else {
				    // strip comment after 1st blank:
				    inbuf += sp;
				    inbuf = inbuf.substring(0, inbuf.indexOf(" "));

				    outbuf = "{\"sym\": \"$" + inbuf + "\",\"rules\": [";
				    if (lineCount > 0) {
				        outbuf = "]}," + outbuf;
				    }
				}
				rtnval = rtnval + outbuf;
				lineCount++;
			}
    		rtnval = rtnval + "]}]";
		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			try {
				if (br != null)br.close();
			} catch (IOException ex) {
				ex.printStackTrace();
			}
		}   
		return rtnval;
    }

    static String parseRuleGrm(String inbuf) {  // output JSON
        char sp = ' ';
        char dquote = '"';
        String sBlank = " ";
        String token;
        String outbuf = "";
        int endidx;
        int i;
        boolean isNumLiteral, isDigit, isOptChar, isOptEnd;
        boolean firstToken = true;
        boolean firstTime = true;
        boolean wasOptChar = true;
        boolean useComma = true;
        boolean numLitFound = false;
        char ch, oldch;
        inbuf = inbuf.trim();
        ch = sp;
        for (i = 0; i < inbuf.length(); i++) {
            // strip repeated blanks, surround [ and ] with blanks
            oldch = ch;
            ch = inbuf.charAt(i);
            if ((ch == sp) && (oldch == sp))
                continue;
            if (ch == '[') {
                outbuf += ch + sBlank;
                ch = sp;
            }
            else if (ch != ']') {
                outbuf += ch;
            }
            else if (oldch == sp) {
                outbuf += ch + sBlank;
                ch = sp;
            }
            else {
                outbuf += sBlank + ch + sBlank;
                ch = sp;
            }
        }
        inbuf = outbuf.trim() + sp;
        outbuf = "";
        while (inbuf.trim().length() > 0) {
            // process each token per loop iter.
            endidx = inbuf.indexOf(" ");
            token = inbuf.substring(0, endidx);
            inbuf = inbuf.substring(endidx + 1, inbuf.length());
            ch = token.charAt(0);
            useComma = true;
            numLitFound = false;
            isDigit = Character.isDigit(ch);
            isNumLiteral = (ch == NUMLITCH);
            isOptChar = (ch == '[');
            isOptEnd = (ch == ']');
            if (isOptChar) {
                firstToken = true;
                wasOptChar = true;
            }
            if (firstToken) {
                firstToken = false;
                if (firstTime && isOptChar) {
                    token = "-1,[-1";
                }
                else if (!isDigit && !isNumLiteral && !isOptChar) {
                    outbuf += "-1,";
                }
                else if (isOptChar) {
                    useComma = false;
                    firstToken = true;
                }
                else if (isNumLiteral) {
                    // numeric literal follows # char.
                    // always first token after [ char.
                    // strip # char.
                    numLitFound = true;
                    token = token.substring(1, token.length());
                }
            }
            if (!isOptChar && !isOptEnd && (!isDigit || !wasOptChar)) {
                token = wrapQuotes(token);
            }
            if (!isOptChar) {
                wasOptChar = false;
            }
            if (numLitFound) {
                token = "-1," + token;
            }
            outbuf += token;
            if (useComma) {
                outbuf += ",";
            }
            outbuf += sp;
            firstTime = false;
        }
        return outbuf;
    }
    
    static String wrapQuotes(String s) {
        return '"' + s + '"';
    }

    static int parsePct(Object obj) {
        long pctLong = 0;
        int pct;
        try {
            pctLong = (Long)obj;
        } catch (ClassCastException e) {
            pctLong = -2;
        }
        pct = (int)pctLong;
        if ((pct >= -1) && (pct <= 100)) {
            // -1 is default pct.
            // all -1 blocks have equal probability
            return pct;
        }
        return -2;  // error parsing percent token
    }

    static void printWordWrap(String inbuf) {
        int col = 0;
        int idx = 0;
        String word;
        String outbuf = "";
        String sp = " ";
        inbuf = inbuf + sp;
        while (inbuf.length() > 0) {
            idx = inbuf.indexOf(sp);
            word = inbuf.substring(0, idx);
            if ((col + idx) <= lineSize) {
                if (col == 0) {
                    outbuf = word;
                }
                else {
                    outbuf = outbuf + sp + word;
                }
                col += idx + 1;
            }
            else {
                System.out.println(outbuf);
                outbuf = word;
                col = idx + 1;
            }
            inbuf = inbuf.substring(idx + 1, inbuf.length());
        }
        System.out.println(outbuf);
    }

    // JSON pretty-printer

    static void printGrm(String sGrammar, Object obj) {
        char dQuote = '"';
        JSONArray array = (JSONArray)obj;
        JSONObject symdef;
        String symname;
        JSONArray rules, rule, optarray;
        int ruleCount, tokenCount;
        int symCount = array.size();
        String linebuf, token, subtoken;
        Object tokobj, subobj;
        int i, j, k, m, n;
        println("[");
        for (i = 0; i < symCount; i++) {
            if (i > 0) {
                println("]},");
            }
            symdef = (JSONObject)array.get(i);
            symname = (String)symdef.get("sym");
            println("{\"sym\": \"" + symname + "\",");
            println("\"rules\": [");
            rules = (JSONArray)symdef.get("rules");
            ruleCount = rules.size();
            for (j = 0; j < ruleCount; j++) {
                rule = (JSONArray)rules.get(j);
                tokenCount = rule.size();
                linebuf = "  [";
                for (k = 0; k < tokenCount; k++) {
                    tokobj = rule.get(k);
                    if (tokobj instanceof Long) {
                        token = "" + (Long)tokobj;
                    }
                    else if (tokobj instanceof String) {
                        token = dQuote + (String)tokobj + dQuote;
                    }
                    else if (tokobj instanceof JSONArray) {
                        token = "[";
                        optarray = (JSONArray)tokobj;
                        n = optarray.size();
                        for (m = 0; m < n; m++) {
                            subobj = optarray.get(m);
                            if (subobj instanceof Long) {
                                subtoken = "" + (Long)subobj;
                            }
                            else if (subobj instanceof String) {
                                subtoken = dQuote + (String)subobj + dQuote;
                            }
                            else {
                                continue;
                            }
                            token += subtoken + ", ";
                        }
                        token += "]";
                    }
                    else {
                        continue;
                    }
                    linebuf += token + ", ";
                }
                linebuf += "],";
                println(linebuf);
            }
        }
        println("]}]");
    }

    static void jsonDemo() {  // this code came from JSON parser demo

      JSONParser parser = new JSONParser();
      String s = "[0,{\"1\":{\"2\":{\"3\":{\"4\":[5,{\"6\":7}]}}}}]";
      try{
         Object obj = parser.parse(s);
         JSONArray array = (JSONArray)obj;
			
         System.out.println("The 2nd element of array");
         System.out.println(array.get(1));
         System.out.println();

         JSONObject obj2 = (JSONObject)array.get(1);
         System.out.println("Field \"1\"");
         System.out.println(obj2.get("1"));    

         s = "{}";
         obj = parser.parse(s);
         System.out.println(obj);

         s = "[5,]";
         obj = parser.parse(s);
         System.out.println(obj);

         s = "[5,,2]";
         obj = parser.parse(s);
         System.out.println(obj);

      }catch(ParseException pe){
		
         System.out.println("position: " + pe.getPosition());
         System.out.println(pe);
      }
    }
}

class GrmObj {  // generate the grammar tree

    private JSONArray grmArray;
    private char SYMCH;
    
    GrmObj(JSONArray array) {
        grmArray = array;
        SYMCH = twirt.SYMCH;
    }
    
    String getSentence() {
        String sentence, s;
        sentence = expandSymDef(0).trim();
        if (sentence.length() == 0)
            return sentence;
        // capitalize 1st letter and append period (.)
        s = sentence.substring(0, 1);
        sentence = s.toUpperCase() + 
            sentence.substring(1, sentence.length()) + ".";
        return sentence;
    }
    
    private String expandSymDef(int idx) {
        JSONObject symdef = (JSONObject)grmArray.get(idx);
        String symval;
        JSONArray rules = (JSONArray)symdef.get("rules");
        JSONArray rule;
        int ruleIdx;
        int ruleCount = rules.size();
        int i, j, k;
        long pct;
        int negCount = 0;

        if (ruleCount == 0)  // whenever "" is returned: error
            return "";
        j = -1;
        for (i = 0; i < ruleCount; i++) {
            // Use the percentage values of each rule (probabilities)
            // to select rule.
            // All -1 rules have equal probability
            rule = (JSONArray)rules.get(i);
            pct = (Long)rule.get(0);
            if (pct < 0) {
                negCount++;
            }
            else if (rollPctDie((int)pct)) {
                j = i;
                break;
            }
        }
        if (j < 0) {
            // no fixed percentages were selected by rolling pct. dice,
            // so process -1 rules
            if (negCount == 0)
                return "";
            j = (int)Math.floor(Math.random() * negCount);
            k = -1;
            for (i = 0; i < ruleCount; i++) {
                rule = (JSONArray)rules.get(i);
                pct = (Long)rule.get(0);
                if (pct < 0) {
                    j--;
                    if (j < 0) {
                        k = i;
                        break;
                    }
                }
            }
            if (k < 0)
                return "";
            j = k;
        }
        ruleIdx = j;
        rule = (JSONArray)rules.get(ruleIdx);
        symval = expandRule(rule);
        return symval;
    }
    
    // expand all non-terminal symbols in rule, return string
    
    private String expandRule(JSONArray rule) {
        Object tokobj;
        JSONArray opt;
        String symval = "";
        String token, subToken;
        int optCount, pct;
        int tokenCount = rule.size();
        int i, m;
        for (i = 1; i < tokenCount; i++) {
            tokobj = rule.get(i);
            if (tokobj instanceof JSONArray) {
                token = "";
                opt = (JSONArray)tokobj;
                optCount = opt.size();
                pct = twirt.parsePct(opt.get(0));
                if (pct == -1) {
                    pct = 50;  // default percentage for optional blk.
                }
                if (rollPctDie(pct)) {
                    for (m = 1; m < optCount; m++) {
                        subToken = (String)opt.get(m);
                        if (isSymToken(subToken)) {
                            subToken = expandToken(subToken);
                        }
                        subToken = subToken.trim();
                        if (subToken.length() > 0) {
                            token = concTrim(token, subToken);
                        }
                    }
                    token = token.trim();
                }
            }
            else {
                token = (String)tokobj;
                token = token.trim();
                if (isSymToken(token)) {
                    token = expandToken(token);
                }
            }
            if (token.length() > 0) {
                symval = concTrim(symval, token);
            }
        }
        return symval;
    }
    
    private boolean rollPctDie(int pct) {
        double p = pct / 100.0;
        if (Math.random() < p)
            return true;
        return false;
    }

    // stick together with space btwn.
    // no space if 2nd str is an apostrophe (')

    String concTrim(String s1, String s2) {
        String sp = " ";
        char apostrophe = '\'';
        if ((s2.length() > 0) && (s2.charAt(0) == apostrophe))
            sp = "";
        String t = s1.trim() + sp + s2.trim();
        t = t.trim();
        return t;
    }

    // all symbol tokens (non-terminal symbols) preceded by SYMCH

    private boolean isSymToken(String token) {
        if (token.length() == 0)
            return false;
        return (token.charAt(0) == SYMCH);
    }

    // look up matching symbol definition in outer array
    
    private int getSymDefIdx(String token) {
        JSONObject symdef;
        String symval;
        int symCount = grmArray.size();
        int i;
        for (i = 0; i < symCount; i++) {
            symdef = (JSONObject)grmArray.get(i);
            symval = (String)symdef.get("sym");
            if (symval.equals(token)) {
                return i;
            }
        }
        return -1;
    }

    // here is where the recursion takes place
    
    private String expandToken(String token) {
        // when debug done, return "" instead of "<...>": OK

        int idx = getSymDefIdx(token);
        String rtnval = "";
        if (idx < 0) {
            rtnval = "";  //"<bad idx>";
        }
        else {
            rtnval = expandSymDef(idx);
            if (rtnval.length() == 0) {
                rtnval = "";  // "<null string>";
            }
        }
        return rtnval;  // no errors
    }
}
